import logging
import re

from ..decorators import _manage_log_level_via_verbosity

logger = logging.getLogger(__name__)


class FieldDomain:
    """Mixin class for methods of field and domain constructs.

    .. versionadded:: (cfdm) 1.9.0.0

    """

    def _apply_masking_constructs(self):
        """Apply masking to metadata constructs in-place.

        Masking is applied to all metadata constructs with data.

        See `Field.apply_masking` or `Domain.apply_masking` for details

        .. versionadded:: (cfdm) 1.9.0.0

        :Returns:

            `None`

        """
        # Apply masking to the metadata constructs
        for c in self.constructs.filter_by_data(todict=True).values():
            c.apply_masking(inplace=True)

    def _construct(
        self,
        _method,
        _constructs_method,
        identities,
        key=False,
        item=False,
        default=ValueError(),
        cached=None,
        **filter_kwargs,
    ):
        """An interface to `Constructs.filter`.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            _method: `str`
                The name of the calling method.

            _constructs_method: `str`
                The name of the corresponding method that can return
                any number of constructs.

            identities: sequence
                As for the *identities* parameter of the calling
                method.

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{cached: optional}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        """
        if cached is not None:
            return cached

        filter_kwargs["todict"] = True

        c = getattr(self, _constructs_method)(*identities, **filter_kwargs)

        # Return construct, or key, or both, or default
        return self._filter_return_construct(
            c,
            key,
            item,
            default,
            _method,
            identities=identities,
            filter_kwargs=filter_kwargs,
        )

    def _get_data_compression_variables(self, component):
        """TODO."""
        out = []
        for construct in self.constructs.filter_by_data(todict=True).values():
            data = construct.get_data(None)
            if data is None:
                continue

            x = getattr(data, f"get_{component}")(None)
            if x is None:
                continue

            out.append(x)

        for construct in self.constructs.filter_by_data(todict=True).values():
            if not construct.has_bounds():
                continue

            data = construct.get_bounds_data(None)
            if data is None:
                continue

            x = getattr(data, f"get_{component}")(None)
            if x is None:
                continue

            out.append(x)

        for construct in self.coordinates(todict=True).values():
            interior_ring = construct.get_interior_ring(None)
            if interior_ring is None:
                continue

            data = interior_ring.get_data(None)
            if data is None:
                continue

            x = getattr(data, f"get_{component}")(None)
            if x is None:
                continue

            out.append(x)

        return out

    def _get_coordinate_geometry_variables(self, component):
        """Return the list of variables for the geometry coordinates.

        :Parameters:

            component: `str`

        :Returns:

            `list'

        """
        out = []
        for construct in self.coordinates(todict=True).values():
            x = getattr(construct, f"get_{component}")(None)
            if x is None:
                continue

            out.append(x)

        return out

    def _filter_return_construct(
        self,
        c,
        key,
        item,
        default,
        _method,
        identities=None,
        filter_kwargs=None,
    ):
        """Return construct, or key, or both, or default.

        :Parameters:

            c: `Constructs` or `dict`

            key: `bool`, optional
                If True then return the selected construct
                identifier. By default the construct itself is
                returned.

            item: `bool`, optional
                If True then return the selected construct identifier
                and the construct itself. By default the construct
                itself is returned. If *key* is True then *item* is
                ignored.

            _method: `str`
                The name of the ultimate calling method.

            identities: `None` or sequence, optional
                The *identity* positional argument passed in by the
                calling function (i.e. by the user). Used for
                informative error messages.

                .. versionadded:: (cfdm) 1.12.0.0

            filter_kwargs: `None` or `dict`, optional
                The filter_kwargs that were passed in by the calling
                function (i.e. by the user). Used for informative
                error messages.

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

                The selected construct, or its identifier if *key* is
                True, or a tuple of both if *item* is True.

        """
        n = len(c)
        if n == 1:
            k, construct = c.popitem()
            if key:
                return k

            if item:
                return k, construct

            return construct

        if default is None:
            return default

        # Create a nice error message
        if identities is None:
            identities = []
        else:
            identities = [f"{i!r}" for i in identities]

        if filter_kwargs is None:
            filter_kwargs = []
        else:
            filter_kwargs.pop("todict", None)
            filter_kwargs = [
                f"{key}={value!r}" for key, value in filter_kwargs.items()
            ]

        args = ", ".join(identities + filter_kwargs)
        construct_type = _method.replace("_", " ")

        return self._default(
            default,
            f"No {construct_type} constructs found with identity {args}",
        )

    def _filter_interface(
        self,
        _ctypes,
        _method,
        identities,
        cached=None,
        construct=False,
        key=False,
        item=False,
        default=ValueError(),
        _identity_config={},
        **filter_kwargs,
    ):
        """An optimised interface to `Constructs.filter`.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            _ctypes: `tuple` of `str`
                The construct types to restrict the selection to.

            _method: `str`
                The name of the calling method.

            identities: `tuple`
                Select constructs that have an identity, defined by
                their `!identities` methods, that matches any of the
                given tuple values.

                {{value match}}

            {{cached: optional}}

            {{key: `bool`, optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            {{item: `bool`, optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            default: optional
                If *construct* is True and there is no unique construct
                then return the value of the *default* parameter.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

            `Constructs` or `dict` or *cached*
                If *cached* is `None`, return the selected
                constructs. If *cached* is not `None`, return
                *cached*.

        """
        if cached is not None:
            return cached

        if not _ctypes:
            kwargs = filter_kwargs
        else:
            # Ensure that filter_by_types is the first filter applied,
            # as it's the cheapest.
            if not (identities or filter_kwargs):
                # This a very common pattern for which calling
                # filter_by_type directly is faster
                if construct:
                    todict = True
                else:
                    todict = False

                c = self.constructs.filter_by_type(*_ctypes, todict=todict)

                if not construct:
                    # Return Constructs or dict
                    return c

                # Return construct, or key, or both, or default
                return self._filter_return_construct(
                    c,
                    key,
                    item,
                    default,
                    _method,
                )

            kwargs = {"filter_by_type": _ctypes}

            if filter_kwargs:
                if "filter_by_type" in filter_kwargs:
                    raise TypeError(
                        f"{self.__class__.__name__}.{_method}() got an "
                        "unexpected keyword argument 'filter_by_type'"
                    )

                kwargs.update(filter_kwargs)

        # Ensure that filter_by_identity is the last filter applied,
        # as it's expensive.
        if filter_kwargs:
            if identities:
                if "filter_by_identity" in filter_kwargs:
                    raise TypeError(
                        f"Can't set {self.__class__.__name__}.{_method}() "
                        "keyword argument 'filter_by_identity' when "
                        "positional *identities arguments are also set"
                    )

                kwargs["filter_by_identity"] = identities
            elif "filter_by_identity" in filter_kwargs:
                kwargs["filter_by_identity"] = filter_kwargs[
                    "filter_by_identity"
                ]
        elif identities:
            kwargs["filter_by_identity"] = identities

        if construct:
            kwargs["todict"] = True

        c = self.constructs.filter(_identity_config=_identity_config, **kwargs)

        if not construct:
            # Return Constructs or dict
            return c

        # Return construct, or key, or both, or default
        return self._filter_return_construct(
            c,
            key,
            item,
            default,
            _method,
            identities=identities,
            filter_kwargs=filter_kwargs,
        )

    def _original_filenames(self, define=None, update=None, clear=False):
        """The names of files containing the original data and metadata.

        {{original filenames}}

        .. versionadded:: (cfdm) 1.10.0.1

        :Parameters:

            {{define: (sequence of) `str`, optional}}

            {{update: (sequence of) `str`, optional}}

            {{clear: `bool` optional}}

        :Returns:

            `set` or `None`
                {{Returns original filenames}}

                If the *define* or *update* parameter is set then
                `None` is returned.

        """
        out = super()._original_filenames(
            define=define, update=update, clear=clear
        )
        if out is None:
            return

        for c in self.constructs.values():
            try:
                # Not all constructs have original filenames
                out.update(c._original_filenames(clear=clear))
            except AttributeError:
                pass

        return out

    def _unique_construct_names(self):
        """Return unique metadata construct names.

        .. versionadded:: (cfdm) 1.7.0

        **Examples**

        >>> f._unique_construct_names()
        {'cellmethod0': 'method:mean',
         'dimensioncoordinate0': 'latitude',
         'dimensioncoordinate1': 'longitude',
         'dimensioncoordinate2': 'time',
         'auxiliarycoordinate0': 'long_name=station id',
         'domainaxis0': 'ncdim%lat',
         'domainaxis1': 'ncdim%lon',
         'domainaxis2': 'key%domainaxis2'}

        """
        key_to_name = {}
        ignore = self.constructs._ignore

        for ctype, d in self.constructs._constructs.items():
            if ctype in ignore:
                continue

            name_to_keys = {}

            for key, construct in d.items():
                name = construct.identity(default="key%" + key)
                name_to_keys.setdefault(name, []).append(key)
                key_to_name[key] = name

            for name, keys in name_to_keys.items():
                if len(keys) <= 1:
                    continue

                for key in keys:
                    found = re.findall(r"\d+$", key)[0]
                    key_to_name[key] = f"{name}{{{found}}}"

        return key_to_name

    def _unique_domain_axis_identities(self):
        """Return unique domain axis construct names.

        .. versionadded:: (cfdm) 1.7.0

        **Examples**

        >>> f._unique_domain_axis_identities()
        {'domainaxis0': 'latitude(5)',
         'domainaxis1': 'longitude(8)',
         'domainaxis2': 'time(1)'}

        """
        key_to_name = {}
        name_to_keys = {}

        for key, value in self.domain_axes(todict=True).items():
            name_size = (
                self.constructs.domain_axis_identity(key),
                value.get_size(""),
            )
            name_to_keys.setdefault(name_size, []).append(key)
            key_to_name[key] = name_size

        for (name, size), keys in name_to_keys.items():
            if len(keys) == 1:
                key_to_name[keys[0]] = f"{name}({size})"
            else:
                for key in keys:
                    found = re.findall(r"\d+$", key)[0]
                    key_to_name[key] = f"{name}{{{found}}}({size})"

        return key_to_name

    def auxiliary_coordinate(
        self,
        *identity,
        default=ValueError(),
        key=False,
        item=False,
        **filter_kwargs,
    ):
        """Select an auxiliary coordinate construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `construct`, `auxiliary_coordinates`

        :Parameters:

            identity: optional
                Select auxiliary coordinate constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all auxiliary
                coordinate constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(1)
        >>> f.auxiliary_coordinate('latitude')
        <AuxiliaryCoordinate: latitude(10, 9) degrees_N>
        >>> f.auxiliary_coordinate('latitude', key=True)
        'auxiliarycoordinate0'
        >>> f.auxiliary_coordinate('latitude', item=True)
        ('auxiliarycoordinate0', <AuxiliaryCoordinate: latitude(10, 9) degrees_N>)

        """
        return self._construct(
            "auxiliary_coordinate",
            "auxiliary_coordinates",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def auxiliary_coordinates(self, *identities, **filter_kwargs):
        """Return auxiliary coordinate constructs.

        Note that ``f.auxiliary_coordinates(*identities,
        **filter_kwargs)`` is equivalent to
        ``f.constructs.filter(filter_by_type=["auxiliary_coordinate"],
        filter_by_identity=identities, **filter_kwargs)``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `auxiliary_coordinate`

        :Parameters:

            identities: optional
                Select auxiliary coordinate constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                If no identities are provided then all auxiliary
                coordinate constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns constructs}}

        **Examples**

        >>> f.auxiliary_coordinates()
        Constructs:
        {}

        >>> f.auxiliary_coordinates()
        Constructs:
        {'auxiliarycoordinate0': <{{repr}}AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
         'auxiliarycoordinate1': <{{repr}}AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
         'auxiliarycoordinate2': <{{repr}}AuxiliaryCoordinate: long_name:Grid latitude name(10) >}

        """
        return self._filter_interface(
            ("auxiliary_coordinate",),
            "auxiliary_coordinates",
            identities,
            **filter_kwargs,
        )

    def del_mesh_id(self, default=ValueError()):
        """Remove the UGRID mesh topology identifier.

        Different field constructs with the same mesh topology
        identifier may be assumed to have domains with a shared UGRID
        mesh topology.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `get_mesh_id`, `has_mesh_id`, `set_mesh_id`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                mesh topology identifier has not been set.

                {{default Exception}}

        :Returns:

                The removed mesh topology identifier.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_mesh_id('df71b85a99894af094411e7cd21c5d68')
        >>> c.has_mesh_id()
        True
        >>> c.get_mesh_id()
        'df71b85a99894af094411e7cd21c5d68'
        >>> c.del_mesh_id()
        'df71b85a99894af094411e7cd21c5d68'
        >>> c.has_mesh_id()
        False
        >>> print(c.del_mesh_id(None))
        None
        >>> print(c.get_mesh_id(None))
        None

        """
        return self._del_component("mesh_id", default=default)

    def domain_topology(
        self,
        *identity,
        default=ValueError(),
        key=False,
        item=False,
        **filter_kwargs,
    ):
        """Select a domain topology construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `construct`, `domain_topologies`

        :Parameters:

            identity: optional
                Select domain topology constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all domain topology
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(8)
        >>> f.domain_topology()
        <{{repr}}DomainTopology: cell:face(3) >
        >>> f.domain_topology('cell:face', key=True)
        'domaintopology0'
        >>> f.domain_topology('cell:face', item=True)
        ('domaintopology0', <{{repr}}DomainTopology: cell:face(3) >)

        """
        return self._construct(
            "domain_topology",
            "domain_topologies",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def domain_topologies(self, *identities, **filter_kwargs):
        """Return domain topology constructs.

        Note that ``f.domain_topologies(*identities,
        **filter_kwargs)`` is equivalent to

        ``f.constructs.filter(filter_by_type=["domain_topology"],
        filter_by_identity=identities, **filter_kwargs)``.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `constructs`, `domain_topology`

        :Parameters:

            identities: optional
                Select domain topology constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                If no identities are provided then all domain topology
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns constructs}}

        **Examples:**

        >>> f = {{package}}.example_{{class_lower}}(8)
        >>> print(f.domain_topologies())
        Constructs:
        {'domaintopology0': <{{repr}}DomainTopology: 'cell:face(3) >}

        """
        return self._filter_interface(
            ("domain_topology",),
            "domain_topologies",
            identities,
            **filter_kwargs,
        )

    def coordinate(
        self,
        *identity,
        default=ValueError(),
        key=False,
        item=False,
        **filter_kwargs,
    ):
        """Select a dimension or auxiliary coordinate construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `construct`, `coordinates`

        :Parameters:

            identity: optional
                Select dimension or auxiliary coordinate constructs
                that have an identity, defined by their `!identities`
                methods, that matches any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all dimension or
                auxiliary coordinate constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(1)
        >>> f.coordinate('latitude')
        <AuxiliaryCoordinate: latitude(10, 9) degrees_N>
        >>> f.coordinate('grid_latitude')
        <DimensionCoordinate: grid_latitude(10) degrees>
        >>> f.coordinate('grid_latitude', key=True)
        'dimensioncoordinate1'
        >>> f.coordinate('grid_latitude', item=True)
        ('dimensioncoordinate1', <DimensionCoordinate: grid_latitude(10) degrees>)

        """
        return self._construct(
            "coordinate",
            "coordinates",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def coordinates(self, *identities, **filter_kwargs):
        """Return dimension and auxiliary coordinate constructs.

        Note that ``f.coordinates(*identities, **filter_kwargs)`` is
        equivalent to
        ``f.constructs.filter(filter_by_type=["dimension_coordinate",
        "auxiliary_coordinate"], filter_by_identity=identities,
        **filter_kwargs)``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `auxiliary_coordinates`, `constructs`,
                     `dimension_coordinates`

        :Parameters:

            identities: optional
                Select dimension and auxiliary coordinate constructs
                that have an identity, defined by their `!identities`
                methods, that matches any of the given values.

                If no identities are provided then all dimension and
                auxiliary coordinate constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns constructs}}

        **Examples**

        >>> f = {{package}}.example_field(7)
        >>> print(f)
        Field: eastward_wind (ncvar%ua)
        -------------------------------
        Data            : eastward_wind(time(3), air_pressure(1), grid_latitude(4), grid_longitude(5)) m s-1
        Cell methods    : time(3): mean
        Dimension coords: time(3) = [1979-05-01 12:00:00, 1979-05-02 12:00:00, 1979-05-03 12:00:00] gregorian
                        : air_pressure(1) = [850.0] hPa
                        : grid_latitude(4) = [0.44, ..., -0.88] degrees
                        : grid_longitude(5) = [-1.18, ..., 0.58] degrees
        Auxiliary coords: latitude(grid_latitude(4), grid_longitude(5)) = [[52.4243, ..., 51.1163]] degrees_north
                        : longitude(grid_latitude(4), grid_longitude(5)) = [[8.0648, ..., 10.9238]] degrees_east
        Coord references: grid_mapping_name:rotated_latitude_longitude
        >>> f.coordinates()
        <Constructs: auxiliary_coordinate(2), dimension_coordinate(4)>
        >>> print(f.coordinates())
        Constructs:
        {'auxiliarycoordinate0': <{{repr}}AuxiliaryCoordinate: latitude(4, 5) degrees_north>,
         'auxiliarycoordinate1': <{{repr}}AuxiliaryCoordinate: longitude(4, 5) degrees_east>,
         'dimensioncoordinate0': <{{repr}}DimensionCoordinate: time(3) days since 1979-1-1 gregorian>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: air_pressure(1) hPa>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: grid_latitude(4) degrees>,
         'dimensioncoordinate3': <{{repr}}DimensionCoordinate: grid_longitude(5) degrees>}

        """
        return self._filter_interface(
            ("dimension_coordinate", "auxiliary_coordinate"),
            "coordinates",
            identities,
            **filter_kwargs,
        )

    def coordinate_reference(
        self,
        *identity,
        default=ValueError(),
        key=False,
        item=False,
        **filter_kwargs,
    ):
        """Return a coordinate reference construct, or its key.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `construct`, `coordinate_references`

        :Parameters:

            identity: optional
                Select coordinate reference constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no identities are provided then all coordinate
                reference constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(1)
        >>> f.coordinate_reference('grid_mapping_name:rotated_latitude_longitude')
        <CoordinateReference: grid_mapping_name:rotated_latitude_longitude>
        >>> f.coordinate_reference('grid_mapping_name:rotated_latitude_longitude', key=True)
        'coordinatereference1'
        >>> f.coordinate_reference('grid_mapping_name:rotated_latitude_longitude', item=True)
        ('coordinatereference1',
         <CoordinateReference: grid_mapping_name:rotated_latitude_longitude>)

        """
        return self._construct(
            "coordinate_reference",
            "coordinate_references",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def coordinate_references(self, *identities, **filter_kwargs):
        """Return coordinate reference constructs.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `coordinate_reference`

        :Parameters:

            identities: optional
                Select coordinate reference constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no identities are provided then all coordinate
                reference constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns constructs}}

        **Examples**

        >>> f.coordinate_references()
        Constructs:
        {}

        >>> f.coordinate_references()
        Constructs:
        {'coordinatereference0': <{{repr}}CoordinateReference: atmosphere_hybrid_height_coordinate>,
         'coordinatereference1': <{{repr}}CoordinateReference: rotated_latitude_longitude>}

        """
        return self._filter_interface(
            ("coordinate_reference",),
            "coordinate_references",
            identities,
            **filter_kwargs,
        )

    def del_construct(self, *identity, default=ValueError(), **filter_kwargs):
        """Remove a metadata construct.

        If a domain axis construct is selected for removal then it
        can't be spanned by any data arrays of the field nor metadata
        constructs, nor be referenced by any cell method
        constructs. However, a domain ancillary construct may be
        removed even if it is referenced by coordinate reference
        construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_construct`, `constructs`, `has_construct`,
                     `set_construct`

        :Parameters:

            identity:
                Select the unique construct that has the identity,
                defined by its `!identities` method, that matches the
                given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                {{value match}}

                {{displayed identity}}

            default: optional
                Return the value of the *default* parameter if the
                data axes have not been set.

                {{default Exception}}

            {{filter_kwargs: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                The removed metadata construct.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> print(f)
        Field: specific_humidity (ncvar%q)
        ----------------------------------
        Data            : specific_humidity(latitude(5), longitude(8)) 1
        Cell methods    : area: mean
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(8) = [22.5, ..., 337.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]
        >>> f.del_construct('time')
        <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >
        >>> f.del_construct('time')
        Traceback (most recent call last):
            ...
        ValueError: Can't find unique construct to remove
        >>> f.del_construct('time', default='No time')
        'No time'
        >>> f.del_construct('dimensioncoordinate1')
        <{{repr}}DimensionCoordinate: longitude(8) degrees_east>
        >>> print(f)
        Field: specific_humidity (ncvar%q)
        ----------------------------------
        Data            : specific_humidity(latitude(5), ncdim%lon(8)) 1
        Cell methods    : area: mean
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north

        """
        key = self.construct_key(*identity, default=None, **filter_kwargs)
        if key is not None:
            return self.constructs._del_construct(key)

        if default is None:
            return default

        return self._default(default, "Can't find unique construct to remove")

    def dimension_coordinate(
        self,
        *identity,
        key=False,
        default=ValueError(),
        item=False,
        **filter_kwargs,
    ):
        """Select a dimension coordinate construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `construct`, `dimension_coordinates`

        :Parameters:

            identity: optional
                Select dimension coordinate constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all dimension
                coordinate constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(1)
        >>> f.dimension_coordinate('grid_latitude')
        <DimensionCoordinate: grid_latitude(10) degrees>
        >>> f.coordinate('grid_latitude', key=True)
        'dimensioncoordinate1'
        >>> f.coordinate('grid_latitude', item=True)
        ('dimensioncoordinate1', <DimensionCoordinate: grid_latitude(10) degrees>)

        """
        return self._construct(
            "dimension_coordinate",
            "dimension_coordinates",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def dimension_coordinates(self, *identities, **filter_kwargs):
        """Return dimension coordinate constructs.

        Note that ``f.dimension_coordinates(*identities,
        **filter_kwargs)`` is equivalent to
        ``f.constructs.filter(filter_by_type=["dimension_coordinate"],
        filter_by_identity=identities, **filter_kwargs)``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `dimension_coordinate`

        :Parameters:

            identities: optional
                Select dimension coordinate constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                If no identities are provided then all dimension
                coordinate constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns constructs}}

        **Examples**

        >>> f.dimension_coordinates()
        Constructs:
        {}

        >>> f.dimension_coordinates()
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: grid_latitude(10) degrees>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: grid_longitude(9) degrees>,
         'dimensioncoordinate3': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >}

        """
        return self._filter_interface(
            ("dimension_coordinate",),
            "dimension_coordinates",
            identities,
            **filter_kwargs,
        )

    def domain_axis(
        self,
        *identity,
        key=False,
        default=ValueError(),
        item=False,
        **filter_kwargs,
    ):
        """Select a domain axis construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `construct`, `domain_axes`

        :Parameters:

            identities: `tuple`, optional
                Select domain axis constructs that have an identity,
                defined by their `!identities` methods, that matches
                any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                Additionally, if for a given ``value``,
                ``f.coordinates(value, filter_by_naxes=(1,))`` returns
                1-d coordinate constructs that all span the same
                domain axis construct then that domain axis construct
                is selected. See `coordinates` for details.

                Additionally, if there is a `Field` data array and a
                value matches the integer position of an array
                dimension, then the corresponding domain axis
                construct is selected.

                If no values are provided then all domain axis
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(1)
        >>> f.domain_axis('domainaxis0')
        <DomainAxis: size(1)>
        >>> f.domain_axis('domainaxis0', key=True)
        'domainaxis0'
        >>> f.domain_axis('domainaxis0', item=True)
        ('domainaxis0', <DomainAxis: size(1)>)

        """
        return self._construct(
            "domain_axis",
            "domain_axes",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def domain_axes(self, *identities, **filter_kwargs):
        """Return domain axis constructs.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `domain_axis`

        :Parameters:

            identities: `tuple`, optional
                Select domain axis constructs that have an identity,
                defined by their `!identities` methods, that matches
                any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                Additionally, if for a given ``value``,
                ``f.coordinates(value, filter_by_naxes=(1,))`` returns
                1-d coordinate constructs that all span the same
                domain axis construct then that domain axis construct
                is selected. See `coordinates` for details.

                Additionally, if there is an associated `Field` data
                array and a value matches the integer position of an
                array dimension, then the corresponding domain axis
                construct is selected.

                If no values are provided then all domain axis
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns constructs}}

        **Examples**

        """
        return self.constructs.domain_axes(*identities, **filter_kwargs)

    def domain_ancillary(
        self,
        *identity,
        default=ValueError(),
        key=False,
        item=False,
        **filter_kwargs,
    ):
        """Select a domain ancillary construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `construct`, `domain_ancillaries`

        :Parameters:

            identity: optional
                Select domain ancillary constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all domain ancillary
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(1)
        >>> f.domain_ancillary('surface_altitude')
        <DomainAncillary: surface_altitude(10, 9) m>
        >>> f.domain_ancillary('surface_altitude', key=True)
        'domainancillary2'
        >>> f.domain_ancillary('surface_altitude', item=True)
        ('domainancillary2', <DomainAncillary: surface_altitude(10, 9) m>)

        """
        return self._construct(
            "domain_ancillary",
            "domain_ancillaries",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def domain_ancillaries(self, *identities, **filter_kwargs):
        """Return domain ancillary constructs.

        Note that ``f.domain_ancillaries(*identities,
        **filter_kwargs)`` is equivalent to
        ``f.constructs.filter(filter_by_type=["domain_ancillary"],
        filter_by_identity=identities, **filter_kwargs)``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `domain_ancillary`

        :Parameters:

            identities: optional
                Select domain ancillary constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns constructs}}

        **Examples**

        >>> f.domain_ancillaries()
        Constructs:
        {}

        >>> f.domain_ancillaries()
        Constructs:
        {'domainancillary0': <{{repr}}DomainAncillary: ncvar%a(1) m>,
         'domainancillary1': <{{repr}}DomainAncillary: ncvar%b(1) >,
         'domainancillary2': <{{repr}}DomainAncillary: surface_altitude(10, 9) m>}

        """
        return self._filter_interface(
            ("domain_ancillary",),
            "domain_ancillaries",
            identities,
            **filter_kwargs,
        )

    def cell_measure(
        self,
        *identity,
        default=ValueError(),
        key=False,
        item=False,
        **filter_kwargs,
    ):
        """Select a cell measure construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `construct`, `cell_measures`

        :Parameters:

            identity: optional
                Select cell measure constructs that have an identity,
                defined by their `!identities` methods, that matches
                any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all cell measure
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(1)
        >>> f.cell_measure()
        <CellMeasure: measure:area(9, 10) km2>
        >>> f.cell_measure('measure:area', key=True)
        'cellmeasure0'
        >>> f.cell_measure('measure:area', item=True)
        ('cellmeasure0', <CellMeasure: measure:area(9, 10) km2>)

        """
        return self._construct(
            "cell_measure",
            "cell_measures",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def cell_measures(self, *identities, **filter_kwargs):
        """Return cell measure constructs.

        ``f.cell_measures(*identities, **filter_kwargs)`` is
        equivalent to
        ``f.constructs.filter(filter_by_type=["cell_measure"],
        filter_by_identity=identities, **filter_kwargs)``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `cell_measure`

        :Parameters:

            identities: optional
                Select cell measure constructs that have an identity,
                defined by their `!identities` methods, that matches
                any of the given values.

                If no identities are provided then all cell measure
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns constructs}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(1)
        >>> print(f.cell_measures())
        Constructs:
        {'cellmeasure0': <{{repr}}CellMeasure: measure%area(9, 10) km2>}

        """
        return self._filter_interface(
            ("cell_measure",), "cell_measures", identities, **filter_kwargs
        )

    def cell_connectivity(
        self,
        *identity,
        default=ValueError(),
        key=False,
        item=False,
        **filter_kwargs,
    ):
        """Select a cell connectivity construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `construct`, `cell_connectivities`

        :Parameters:

            identity: optional
                Select cell connectivity constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all cell connectivity
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{key: `bool`, optional}}

            {{item: `bool`, optional}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.
                {{default Exception}}

            {{filter_kwargs: optional}}

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(8)
        >>> f.cell_connectivity()
        <{{repr}}CellConnectivity: connectivity:edge(3) >
        >>> f.cell_connectivity('connectivity:edge', key=True)
        'cellconnectivity0'
        >>> f.cell_connectivity('connectivity:edge', item=True)
        ('cellconnectivity0', <{{repr}}CellConnectivity: connectivity:edge(3) >)

        """
        return self._construct(
            "cell_connectivity",
            "cell_connectivities",
            identity,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def cell_connectivities(self, *identities, **filter_kwargs):
        """Select cell connectivity constructs.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `construct`, `cell_connectivity`

        Note that ``f.cell_connectivities(*identities, **filter_kwargs)``
        is equivalent to
        ``f.constructs.filter(filter_by_type=["cell_connectivity"],
        filter_by_identity=identities, **filter_kwargs)``.


        :Parameters:

            identities: optional
                Select cell connectivity constructs that have an
                identity, defined by their `!identities` methods, that
                matches any of the given values.

                If no identities are provided then all cell
                connectivity constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

        :Returns:

                {{Returns constructs}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(8)
        >>> print(f.cell_connectivities())
        Constructs:
        {'cellconnectivity0': <{{repr}}CellConnectivity: connectivity:edge(3) >}

        """
        return self._filter_interface(
            ("cell_connectivity",),
            "cell_connectivities",
            identities,
            **filter_kwargs,
        )

    def construct(
        self,
        *identity,
        default=ValueError(),
        key=False,
        item=False,
        **filter_kwargs,
    ):
        """Return a metadata construct.

        {{unique construct}}

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `del_construct`, `get_construct`,
                     `has_construct`, `set_construct`

        :Parameters:

            identity: optional
                Select constructs that have an identity, defined by
                their `!identities` methods, that matches any of the
                given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all constructs are
                selected.

                {{value match}}

                {{displayed identity}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{key: `bool`, optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            {{item: `bool`, optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            {{filter_kwargs: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns construct}}

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(0)

        Select the construct that has the "standard_name" property of
        'latitude':

        >>> print(f.constructs.filter_by_property(standard_name=None))
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: long_name=time(1) days since 2018-12-01 >}
        >>> f.construct('latitude')
        <{{repr}}DimensionCoordinate: latitude(5) degrees_north>

        Attempt to select a unique construct whose "standard_name"
        start with the letter 'l':

        >>> import re
        >>> f.construct(re.compile('^l'))
        Traceback
            ...
        ValueError: {{class}}.construct() can't return 2 constructs
        >>> f.construct(re.compile('^l'), default='no unique construct')
        'no unique construct'

        """
        return self._filter_interface(
            (),
            "construct",
            identity,
            construct=True,
            key=key,
            item=item,
            default=default,
            **filter_kwargs,
        )

    def construct_item(self, *identity, default=ValueError(), **filter_kwargs):
        """Return a metadata construct and its identifier.

        {{unique construct}}

        ``f.construct_item(*args, **kwargs)`` is an alias for
        ``f.construct(*args, item=True, **kwargs)``. See `construct`
        for details.

        .. versionadded:: (cfdm) 1.8.9.0

        .. seealso:: `construct`, `construct_key`

        :Parameters:

            identity: optional
                Select constructs that have an identity, defined by
                their `!identities` methods, that matches any of the
                given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all constructs are
                selected.

                {{value match}}

                {{displayed identity}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `tuple`
                The selected construct and its construct identifier.

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(0)

        Select the construct that has the "standard_name" property of
        'latitude':

        >>> print(f.constructs.filter_by_property(standard_name=None))
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: long_name=time(1) days since 2018-12-01 >}
        >>> f.construct_item('latitude')
        ('dimensioncoordinate1', <{{repr}}DimensionCoordinate: latitude(5) degrees_north>)

        Attempt to select a unique construct whose "standard_name"
        start with the letter 'l':

        >>> import re
        >>> f.construct(re.compile('^l'))
        Traceback
            ...
        ValueError: {{class}}.construct() can't return 2 constructs
        >>> f.construct(re.compile('^l'), default='no unique construct')
        'no unique construct'

        """
        return self._filter_interface(
            (),
            "construct",
            identity,
            construct=True,
            key=False,
            item=True,
            default=default,
            **filter_kwargs,
        )

    def construct_key(self, *identity, default=ValueError(), **filter_kwargs):
        """Return the identifier of a metadata construct.

        {{unique construct}}

        ``f.construct_key(*args, **kwargs)`` is an alias for
        ``f.construct(*args, itekey=True, **kwargs)``. See `construct`
        for details.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `construct`, `construct_item`

        :Parameters:

            identity: optional
                Select constructs that have an identity, defined by
                their `!identities` methods, that matches any of the
                given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                If no values are provided then all constructs are
                selected.

                {{value match}}

                {{displayed identity}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `str`
                The identifier of the selected construct.

        **Examples**

        >>> f = {{package}}.example_{{class_lower}}(0)

        Select the construct that has the "standard_name" property of
        'latitude':

        >>> print(f.constructs.filter_by_property(standard_name=None))
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: long_name=time(1) days since 2018-12-01 >}
        >>> f.construct_key('latitude')
        'dimensioncoordinate1'

        Attempt to select a unique construct whose "standard_name"
        start with the letter 'l':

        >>> import re
        >>> f.construct(re.compile('^l'))
        Traceback
            ...
        ValueError: {{class}}.construct() can't return 2 constructs
        >>> f.construct(re.compile('^l'), default='no unique construct')
        'no unique construct'

        """
        return self._filter_interface(
            (),
            "construct",
            identity,
            construct=True,
            key=True,
            item=False,
            default=default,
            **filter_kwargs,
        )

    def domain_axis_key(
        self, *identity, default=ValueError(), **filter_kwargs
    ):
        """Returns the domain axis key spanned by 1-d coordinates.

        Specifically, returns the key of the domain axis construct that
        is spanned by 1-d coordinate constructs.

        :Parameters:

            identity: optional
                Select the 1-d dimension coordinate constructs that
                have an identity, defined by their `!identities`
                methods, that matches any of the given values.

                If no values are provided then all 1-d dimension
                coordinate constructs are selected.

                {{value match}}

                {{displayed identity}}

            default: optional
                Return the value of the *default* parameter if there
                is no unique construct.

                {{default Exception}}

            {{filter_kwargs: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `str`
                The key of the domain axis construct that is spanned by
                the data of the selected 1-d coordinate constructs.

        **Examples**

        >>> print(f.constructs())
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: time(1) days since 1964-01-21 00:00:00 >,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: pressure(23) mbar>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: latitude(160) degrees_north>,
         'dimensioncoordinate3': <{{repr}}DimensionCoordinate: longitude(320) degrees_east>,
         'domainaxis0': <{{repr}}DomainAxis: size(1)>,
         'domainaxis1': <{{repr}}DomainAxis: size(23)>,
         'domainaxis2': <{{repr}}DomainAxis: size(160)>,
         'domainaxis3': <{{repr}}DomainAxis: size(320)>}
        >>> f.domain.domain_axis_key('time')
        'domainaxis0'
        >>> f.domain.domain_axis_key('longitude')
        'domainaxis3'

        """
        filter_kwargs["filter_by_naxes"] = (1,)
        filter_kwargs["todict"] = True

        # Select 1-d coordinate constructs with the given identity
        c = self.coordinates(*identity, **filter_kwargs)
        if not c:
            if default is None:
                return default

            return self._default(
                default, "There are no 1-d coordinate constructs"
            )

        data_axes = self.constructs.data_axes()
        domain_axes = self.domain_axes(todict=True)

        keys = []
        for ckey, coord in c.items():
            axes = data_axes.get(ckey)
            if axes is None:
                continue

            key = axes[0]
            if key in domain_axes:
                keys.append(key)

        keys = set(keys)

        if len(keys) == 1:
            return keys.pop()

        if default is None:
            return default

        if not keys:
            return self._default(
                default,
                "Some selected 1-d coordinate constructs "
                "have not been assigned a domain axis construct",
            )

        return self._default(
            default,
            "The selected 1-d coordinate constructs "
            f"span multiple domain axes: {keys!r}",
        )

    @_manage_log_level_via_verbosity
    def equals(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_data_type=False,
        ignore_fill_value=False,
        ignore_properties=None,
        ignore_compression=True,
        ignore_type=False,
    ):
        """Whether two constructs are the same.

        Equality is strict by default. This means that for two
        constructs to be considered equal they must have corresponding
        metadata constructs and for each pair of constructs:

        * the same descriptive properties must be present, with the
          same values and data types, and vector-valued properties
          must also have same the size and be element-wise equal (see
          the *ignore_properties* and *ignore_data_type* parameters),
          and

        ..

        * if there are data arrays then they must have same shape and
          data type, the same missing data mask, and be element-wise
          equal (see the *ignore_data_type* parameter).

        {{equals tolerance}}

        {{equals compression}}

        Any type of object may be tested but, in general, equality is
        only possible with another field construct, or a subclass of
        one. See the *ignore_type* parameter.

        {{equals netCDF}}

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            other:
                The object to compare for equality.

            {{atol: number, optional}}

            {{rtol: number, optional}}

            {{ignore_fill_value: `bool`, optional}}

            ignore_properties: (sequence of) `str`, optional
                The names of properties of the construct (not the
                metadata constructs) to omit from the comparison. Note
                that the ``Conventions`` property is always omitted.

            {{ignore_data_type: `bool`, optional}}

            {{ignore_compression: `bool`, optional}}

            {{ignore_type: `bool`, optional}}

            {{verbose: `int` or `str` or `None`, optional}}

        :Returns:

            `bool`
                Whether the two constructs are equal.

        **Examples**

        >>> f.equals(f)
        True
        >>> f.equals(f.copy())
        True
        >>> f.equals(f[...])
        True
        >>> f.equals('a string, not a construct')
        False

        >>> g = f.copy()
        >>> g.set_property('foo', 'bar')
        >>> f.equals(g)
        False
        >>> f.equals(g, verbose=3)
        {{class}}: Non-common property name: foo
        {{class}}: Different properties
        False

        """
        # Check the properties and data
        if not ignore_properties:
            ignore_properties = ("Conventions",)
        elif isinstance(ignore_properties, str):
            ignore_properties = (ignore_properties, "Conventions")
        else:
            ignore_properties = tuple(ignore_properties) + ("Conventions",)

        if not super().equals(
            other,
            rtol=rtol,
            atol=atol,
            verbose=verbose,
            ignore_data_type=ignore_data_type,
            ignore_fill_value=ignore_fill_value,
            ignore_properties=ignore_properties,
            ignore_compression=ignore_compression,
            ignore_type=ignore_type,
        ):
            return False

        # Check the constructs
        if not self.constructs.equals(
            other.constructs,
            rtol=rtol,
            atol=atol,
            verbose=verbose,
            ignore_data_type=ignore_data_type,
            ignore_fill_value=ignore_fill_value,
            ignore_compression=ignore_compression,
            _ignore_type=False,
        ):
            logger.info(
                f"{self.__class__.__name__}: Different metadata constructs"
            )
            return False

        return True

    def get_mesh_id(self, default=ValueError()):
        """Return the UGRID mesh topology identifier.

        Different field constructs with the same mesh topology
        identifier may be assumed to have domains with a shared UGRID
        mesh topology.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `del_mesh_id`, `has_mesh_id`, `set_mesh_id`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                mesh topology identifier has not been set.

                {{default Exception}}

        :Returns:

                The value of the mesh topology identifier.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_mesh_id('df71b85a99894af094411e7cd21c5d68')
        >>> c.has_mesh_id()
        True
        >>> c.get_mesh_id()
        'df71b85a99894af094411e7cd21c5d68'
        >>> c.del_mesh_id()
        'df71b85a99894af094411e7cd21c5d68'
        >>> c.has_mesh_id()
        False
        >>> print(c.del_mesh_id(None))
        None
        >>> print(c.get_mesh_id(None))
        None

        """
        return self._get_component("mesh_id", default=default)

    def has_construct(self, *identity, **filter_kwargs):
        """Whether a unique metadata construct exists.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_construct`, `constructs`, `has_construct`,
                     `set_construct`

        :Parameters:

            identity:
                Select the unique construct that has the identity,
                defined by its `!identities` method, that matches the
                given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                {{value match}}

                {{displayed identity}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{filter_kwargs: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                Whether or not a unique construct can be identified.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> print(f)
        Field: specific_humidity (ncvar%q)
        ----------------------------------
        Data            : specific_humidity(latitude(5), longitude(8)) 1
        Cell methods    : area: mean
        Dimension coords: latitude(5) = [-75.0, ..., 75.0] degrees_north
                        : longitude(8) = [22.5, ..., 337.5] degrees_east
                        : time(1) = [2019-01-01 00:00:00]
        >>> f.has_construct('time')
        True
        >>> f.del_construct('altitude')
        False

        """
        return bool(
            self.construct(*identity, default=None, **filter_kwargs)
            is not None
        )

    def has_geometry(self):
        """Return whether or not any coordinates have cell geometries.

        .. versionadded:: (cfdm) 1.8.0

        :Returns:

            `bool`
                True if there are geometries, otherwise False.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.has_geometry()
        False

        """
        for c in self.constructs.filter_by_type(
            "auxiliary_coordinate",
            "dimension_coordinate",
            "domain_ancillary",
            todict=True,
        ).values():
            if c.has_geometry():
                return True

        return False

    def has_mesh_id(self):
        """Whether the UGRID mesh topology identifier has been set.

        Different field constructs with the same mesh topology
        identifier may be assumed to have domains with a shared UGRID
        mesh topology.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `del_mesh_id`, `get_mesh_id`, `set_mesh_id`

        :Returns:

            `bool`
                True if the mesh topology identifier has been set,
                otherwise False.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_mesh_id('df71b85a99894af094411e7cd21c5d68')
        >>> c.has_mesh_id()
        True
        >>> c.get_mesh_id()
        'df71b85a99894af094411e7cd21c5d68'
        >>> c.del_mesh_id()
        'df71b85a99894af094411e7cd21c5d68'
        >>> c.has_mesh_id()
        False
        >>> print(c.del_mesh_id(None))
        None
        >>> print(c.get_mesh_id(None))
        None

        """
        return self._has_component("mesh_id")

    def set_mesh_id(self, mesh_id):
        """Set a UGRID mesh topology identifier.

        Different field constructs with the same mesh topology
        identifier may be assumed to have domains with a shared UGRID
        mesh topology.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `del_mesh_id`, `get_mesh_id`, `has_mesh_id`

        :Parameters:

            mesh_id: `str`
                The value for the mesh topology identifier.

        :Returns:

             `None`

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_mesh_id('df71b85a99894af094411e7cd21c5d68')
        >>> c.has_mesh_id()
        True
        >>> c.get_mesh_id()
        'df71b85a99894af094411e7cd21c5d68'
        >>> c.del_mesh_id()
        'df71b85a99894af094411e7cd21c5d68'
        >>> c.has_mesh_id()
        False
        >>> print(c.del_mesh_id(None))
        None
        >>> print(c.get_mesh_id(None))
        None

        """
        return self._set_component("mesh_id", mesh_id, copy=False)



import datetime
import faulthandler
import os
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class CellMethodTest(unittest.TestCase):
    """Unit test for the CellMethod class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or calls (those
        # without a 'verbose' option to do the same) e.g. to debug them, wrap
        # them (for methods, start-to-end internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        self.filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
        )
        f = cfdm.read(self.filename)
        self.assertEqual(len(f), 1, f"f={f!r}")
        self.f = f[0]

    def test_CellMethod__repr__str__dump_construct_type(self):
        """Test all means of CellMethod inspection."""
        f = self.f

        for c in f.cell_methods().values():
            _ = repr(c)
            _ = str(c)
            self.assertIsInstance(c.dump(display=False), str)
            self.assertEqual(c.construct_type, "cell_method")

    def test_CellMethod(self):
        """Test CellMethod equality, identity and sorting methods."""
        f = self.f

        # ------------------------------------------------------------
        # Equals and identities
        # ------------------------------------------------------------
        for c in f.cell_methods().values():
            d = c.copy()
            self.assertTrue(c.equals(c, verbose=3))
            self.assertTrue(c.equals(d, verbose=3))
            self.assertTrue(d.equals(c, verbose=3))
            self.assertEqual(c.identity(), "method:" + c.get_method())
            self.assertEqual(c.identities(), ["method:" + c.get_method()])

        # ------------------------------------------------------------
        # Sorted
        # ------------------------------------------------------------
        c = cfdm.CellMethod(
            method="minimum", axes=["B", "A"], qualifiers={"interval": [1, 2]}
        )

        d = cfdm.CellMethod(
            method="minimum", axes=["A", "B"], qualifiers={"interval": [2, 1]}
        )

        self.assertTrue(d.equals(c.sorted(), verbose=3))

        c = cfdm.CellMethod(
            method="minimum", axes=["B", "A"], qualifiers={"interval": [3]}
        )

        d = cfdm.CellMethod(
            method="minimum", axes=["A", "B"], qualifiers={"interval": [3]}
        )

        self.assertTrue(d.equals(c.sorted(), verbose=3))

        c = cfdm.CellMethod(
            method="minimum", axes=["area"], qualifiers={"interval": [3]}
        )

        d = cfdm.CellMethod(
            method="minimum", axes=["area"], qualifiers={"interval": [3]}
        )

        self.assertTrue(d.equals(c.sorted(), verbose=3))

        # Init
        c = cfdm.CellMethod(source="qwerty")

    def test_CellMethod_axes(self):
        """Test the axes access and (un)setting CellMethod methods."""
        f = cfdm.CellMethod()

        self.assertFalse(f.has_axes())
        self.assertIsNone(f.get_axes(None))
        self.assertIsNone(f.set_axes(["time"]))
        self.assertTrue(f.has_axes())
        self.assertEqual(f.get_axes(), ("time",))
        self.assertEqual(f.del_axes(), ("time",))
        self.assertIsNone(f.del_axes(None))

    def test_CellMethod_method(self):
        """Test the method access and (un)setting CellMethod methods."""
        f = cfdm.CellMethod()

        self.assertFalse(f.has_method())
        self.assertIsNone(f.get_method(None))
        self.assertIsNone(f.set_method("mean"))
        self.assertTrue(f.has_method())
        self.assertEqual(f.get_method(), "mean")
        self.assertEqual(f.del_method(), "mean")
        self.assertIsNone(f.del_method(None))

    def test_CellMethod_qualifier(self):
        """Test qualifier access and (un)setting CellMethod methods."""
        f = cfdm.CellMethod()

        self.assertEqual(f.qualifiers(), {})
        self.assertFalse(f.has_qualifier("within"))
        self.assertIsNone(f.get_qualifier("within", None))
        self.assertIsNone(f.set_qualifier("within", "years"))
        self.assertEqual(f.qualifiers(), {"within": "years"})
        self.assertTrue(f.has_qualifier("within"))
        self.assertEqual(f.get_qualifier("within"), "years")
        self.assertEqual(f.del_qualifier("within"), "years")
        self.assertIsNone(f.del_qualifier("within", None))
        self.assertEqual(f.qualifiers(), {})


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



from itertools import accumulate
from numbers import Number

from ..subarray import RaggedSubarray
from .compressedarray import CompressedArray


class RaggedArray(CompressedArray):
    """An underlying ragged array.

    See CF section 9 "Discrete Sampling Geometries"

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def __new__(cls, *args, **kwargs):
        """Store subarray classes.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        instance = super().__new__(cls)

        instance._Subarray = {
            "ragged contiguous": RaggedSubarray,
            "ragged indexed": RaggedSubarray,
            "ragged indexed contiguous": RaggedSubarray,
        }

        return instance

    def __init__(
        self,
        compressed_array=None,
        shape=None,
        compressed_dimensions=None,
        count_variable=None,
        index_variable=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            compressed_array: array_like
                The compressed data.

            shape: `tuple`
                The shape of the uncompressed array.

            compressed_dimensions: `dict`
                Mapping of dimensions of the compressed array to their
                corresponding dimensions in the uncompressed
                array. Compressed array dimensions that are not
                compressed must be omitted from the mapping.

                *Parameter example:*
                  ``{0: (0, 1)}``

                *Parameter example:*
                  ``{0: (0, 1, 2)}``

            count_variable: `Count`, optional
                A count variable for uncompressing the data,
                corresponding to a CF-netCDF count variable, if
                required by the decompression method.

            index_variable: `Index`, optional
                An index variable for uncompressing the data,
                corresponding to a CF-netCDF count variable, if
                required by the decompression method.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        if count_variable is not None:
            if index_variable is not None:
                compression_type = "ragged indexed contiguous"
            else:
                compression_type = "ragged contiguous"
        elif index_variable is not None:
            compression_type = "ragged indexed"
        else:
            compression_type = None

        super().__init__(
            compressed_array=compressed_array,
            shape=shape,
            compressed_dimensions=compressed_dimensions,
            compression_type=compression_type,
            source=source,
            copy=copy,
        )

        if source is not None:
            try:
                index_variable = source.get_index(None)
            except AttributeError:
                index_variable = None

            try:
                count_variable = source.get_count(None)
            except AttributeError:
                count_variable = None

        if index_variable is not None:
            self._set_component("index_variable", index_variable, copy=copy)

        if count_variable is not None:
            self._set_component("count_variable", count_variable, copy=copy)

    def _uncompressed_descriptors(self, u_dims, shapes):
        """Create descriptors of uncompressed subarrays.

        Returns the location of each subarray, the indices of the
        uncompressed array that correspond to each subarray, and the
        shape of each uncompressed subarray.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `subarray`

        :Parameters:

            u_dims: sequence of `int`
                The positions of the uncompressed dimensions that have
                been compressed.

            shapes: sequence of `tuple`
                The subarray shapes along each uncompressed
                dimension. It is assumed that they have been output by
                `subarray_shapes`.

        :Returns:

            3-`tuple` of `list`
                The location, shape, and index descriptors.

        """
        locations = []
        u_shapes = []
        u_indices = []
        for d, (size, c) in enumerate(zip(self.shape, shapes)):
            if d in u_dims[:-1]:
                locations.append([i for i in range(len(c))])
                u_shapes.append(c)
                u_indices.append([slice(i, i + 1) for i in range(len(c))])
            elif d == u_dims[-1]:
                locations.append((0,))
                u_shapes.append((size,))
                u_indices.append((slice(0, size),))
            else:
                locations.append([i for i in range(len(c))])
                u_shapes.append(c)

                c = tuple(accumulate((0,) + c))
                u_indices.append([slice(i, j) for i, j in zip(c[:-1], c[1:])])

        return locations, u_shapes, u_indices

    @property
    def dtype(self):
        """Data-type of the uncompressed data."""
        return self.source().dtype

    def get_count(self, default=ValueError()):
        """Return the count variable for the compressed array.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `Count`

        """
        out = self._get_component("count_variable", None)
        if out is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__!r} has no count variable"
            )

        return out

    def get_index(self, default=ValueError()):
        """Return the index variable for the compressed array.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `Index`

        """
        out = self._get_component("index_variable", None)
        if out is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__!r} has no index variable"
            )

        return out

    def subarray_shapes(self, shapes):
        """Create the subarray shapes along each uncompressed dimension.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `subarray`

        :Parameters:

            {{subarray_shapes chunks: `int`, sequence, `dict`, or `str`, optional}}

        :Returns:

            `list`
                The subarray sizes along each uncompressed dimension.

        **Examples**

        >>> a.shape
        (2, 3, 4)
        >>> a.compressed_dimensions()
        {0: (0, 1)}
        >>> a.subarray_shapes(-1)
        [(1, 1), (3,), (4,)]
        >>> a.subarray_shapes("auto")
        [(1, 1), (3,), "auto"]
        >>> a.subarray_shapes(2)
        [(1, 1), (3,), 2]
        >>> a.subarray_shapes("60B")
        [(1, 1), (3,), "60B"]
        >>> a.subarray_shapes((None, None, 2))
        [(1, 1), (3,), 2]
        >>> a.subarray_shapes((None, None, (1, 3)))
        [(1, 1), (3,), (1, 3)]
        >>> a.subarray_shapes((None, None, "auto"))
        [(1, 1), (3,), "auto"]
        >>> a.subarray_shapes((None, None, "60B"))
        [(1, 1), (3,), "60B"]
        >>> a.subarray_shapes({2: (1, 3)})
        [(1, 1), (3,), (1, 3)]

        >>> import dask.array as da
        >>> da.core.normalize_chunks(
        ...   a.subarray_shapes("auto"), shape=a.shape, dtype=a.dtype
        ... )
        [(1, 1), (3,), (4,)]
        >>> da.core.normalize_chunks(
        ...   a.subarray_shapes(2, shape=a.shape, dtype=a.dtype
        ... )
        [(1, 1), (3,), (2, 2)]

        >>> a.shape
        (2, 3, 3, 4)
        >>> a.compressed_dimensions()
        {0: (0, 1, 2)}
        >>> a.subarray_shapes(-1)
        [(1, 1), (1, 1, 1), (3,), (4,)]
        >>> a.subarray_shapes("auto")
        [(1, 1), (1, 1, 1), (3,), "auto"]
         >>> a.subarray_shapes(2)
        [(1, 1), (1, 1, 1), (3,), 2]
         >>> a.subarray_shapes("60B")
        [(1, 1), (1, 1, 1), (3,), "60B"]
        >>> a.subarray_shapes((None, None, None, 2))
        [(1, 1), (1, 1, 1), (3,), 2]
        >>> a.subarray_shapes((None, None, None, (1, 3)))
        [(1, 1), (1, 1, 1), (3,), (1, 3)]
        >>> a.subarray_shapes((None, None, None, "auto"))
        [(1, 1), (1, 1, 1), (3,), "auto"]
        >>> a.subarray_shapes((None, None, None, "60B"))
        [(1, 1), (1, 1, 1), (3,), "60B"]
        >>> a.subarray_shapes({3: (1, 3)})
        [(1, 1), (1, 1, 1), (3,), (1, 3)]

        """
        u_dims = self.get_compressed_axes()

        uncompressed_shape = self.shape

        if shapes == -1:
            return [
                (1,) * size if i in u_dims[:-1] else (size,)
                for i, size in enumerate(uncompressed_shape)
            ]

        if isinstance(shapes, (str, Number)):
            shapes = [
                (1,) * size if i in u_dims[:-1] else shapes
                for i, size in enumerate(uncompressed_shape)
            ]
            shapes[u_dims[-1]] = (uncompressed_shape[u_dims[-1]],)
            return shapes

        if isinstance(shapes, dict):
            shapes = [
                shapes[i] if i in shapes else None for i in range(self.ndim)
            ]
        elif len(shapes) != self.ndim:
            raise ValueError(
                f"Wrong number of 'shapes' elements in {shapes}: "
                f"Got {len(shapes)}, expected {self.ndim}"
            )

        # chunks is a sequence
        shapes = list(shapes)
        shapes[u_dims[0] : u_dims[-1]] = [
            (1,) * size for size in uncompressed_shape[u_dims[0] : u_dims[-1]]
        ]
        shapes[u_dims[-1]] = (uncompressed_shape[u_dims[-1]],)
        return shapes

    def to_memory(self):
        """Bring data on disk into memory.

        There is no change to data that is already in memory.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `{{class}}`
                A copy of the array with all of its data in memory.

        """
        a = super().to_memory()

        count = a.get_count(None)
        if count is not None:
            try:
                a._set_component(
                    "count_variable", count.to_memory(), copy=False
                )
            except AttributeError:
                pass

        index = a.get_index(None)
        if index is not None:
            try:
                a._set_component(
                    "index_variable", index.to_memory(), copy=False
                )
            except AttributeError:
                pass

        return a



import datetime
import faulthandler
import os
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class AuxiliaryCoordinateTest(unittest.TestCase):
    """Unit test for the AuxiliaryCoordinate class."""

    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
    )

    aux1 = cfdm.AuxiliaryCoordinate()
    aux1.standard_name = "latitude"
    a = np.array(
        [
            -30,
            -23.5,
            -17.8123,
            -11.3345,
            -0.7,
            -0.2,
            0,
            0.2,
            0.7,
            11.30003,
            17.8678678,
            23.5,
            30,
        ]
    )
    aux1.set_data(cfdm.Data(a, "degrees_north"))
    bounds = cfdm.Bounds()
    b = np.empty(a.shape + (2,))
    b[:, 0] = a - 0.1
    b[:, 1] = a + 0.1
    bounds.set_data(cfdm.Data(b))
    aux1.set_bounds(bounds)

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_AuxiliaryCoordinate__repr__str__dump(self):
        """Test all means of AuxiliaryCoordinate inspection."""
        f = cfdm.read(self.filename, verbose=1)[0]
        x = f.auxiliary_coordinates("latitude").value()

        repr(x)
        str(x)
        self.assertIsInstance(x.dump(display=False), str)
        self.assertIsInstance(x.dump(display=False, _title=None), str)

    def test_AuxiliaryCoordinate_bounds(self):
        """Test the bounds keyword argument to AuxiliaryCoordinate."""
        f = cfdm.read(self.filename)[0]

        d = f.dimension_coordinates("grid_longitude").value()
        cfdm.AuxiliaryCoordinate(bounds=d)

    def test_AuxiliaryCoordinate_properties(self):
        """Test the property access methods of AuxiliaryCoordinate."""
        f = cfdm.read(self.filename)[0]
        x = f.auxiliary_coordinates("latitude").value()

        x.set_property("long_name", "qwerty")

        self.assertEqual(x.get_property("long_name"), "qwerty")
        self.assertEqual(x.del_property("long_name"), "qwerty")
        self.assertIsNone(x.get_property("long_name", None))
        self.assertIsNone(x.del_property("long_name", None))

    def test_AuxiliaryCoordinate_source(self):
        """Test the source keyword argument to AuxiliaryCoordinate."""
        f = cfdm.read(self.filename)[0]
        d = f.dimension_coordinates("grid_longitude").value()
        cfdm.AuxiliaryCoordinate(source=d)

    def test_AuxiliaryCoordinate_insert_dimension(self):
        """Test the `insert_dimension` method of AuxiliaryCoordinate."""
        f = cfdm.read(self.filename)[0]
        d = f.dimension_coordinates("grid_longitude").value()
        x = cfdm.AuxiliaryCoordinate(source=d)

        self.assertEqual(x.shape, (9,))
        self.assertEqual(x.bounds.shape, (9, 2))

        y = x.insert_dimension(0)
        self.assertEqual(y.shape, (1, 9))
        self.assertEqual(y.bounds.shape, (1, 9, 2), y.bounds.shape)

        x.insert_dimension(-1, inplace=True)
        self.assertEqual(x.shape, (9, 1))
        self.assertEqual(x.bounds.shape, (9, 1, 2), x.bounds.shape)

    def test_AuxiliaryCoordinate_transpose(self):
        """Test the transpose method of AuxiliaryCoordinate."""
        f = cfdm.read(self.filename)[0]
        x = f.auxiliary_coordinates("longitude").value()

        bounds = cfdm.Bounds(
            data=cfdm.Data(np.arange(9 * 10 * 4).reshape(9, 10, 4))
        )
        x.set_bounds(bounds)

        self.assertEqual(x.shape, (9, 10))
        self.assertEqual(x.bounds.shape, (9, 10, 4))

        y = x.transpose()
        self.assertEqual(y.shape, (10, 9))
        self.assertEqual(y.bounds.shape, (10, 9, 4), y.bounds.shape)

        x.transpose([1, 0], inplace=True)
        self.assertEqual(x.shape, (10, 9))
        self.assertEqual(x.bounds.shape, (10, 9, 4), x.bounds.shape)

    def test_AuxiliaryCoordinate_squeeze(self):
        """Test the squeeze method of AuxiliaryCoordinate."""
        f = cfdm.read(self.filename)[0]
        x = f.auxiliary_coordinates("longitude").value()

        bounds = cfdm.Bounds(
            data=cfdm.Data(np.arange(9 * 10 * 4).reshape(9, 10, 4))
        )
        x.set_bounds(bounds)
        x.insert_dimension(1, inplace=True)
        x.insert_dimension(0, inplace=True)

        self.assertEqual(x.shape, (1, 9, 1, 10))
        self.assertEqual(x.bounds.shape, (1, 9, 1, 10, 4))

        y = x.squeeze()
        self.assertEqual(y.shape, (9, 10))
        self.assertEqual(y.bounds.shape, (9, 10, 4), y.bounds.shape)

        x.squeeze(2, inplace=True)
        self.assertEqual(x.shape, (1, 9, 10))
        self.assertEqual(x.bounds.shape, (1, 9, 10, 4), x.bounds.shape)

    def test_AuxiliaryCoordinate_interior_ring(self):
        """Test the interior ring access AuxiliaryCoordinate methods."""
        c = cfdm.AuxiliaryCoordinate()

        i = cfdm.InteriorRing(data=cfdm.Data(np.arange(10).reshape(5, 2)))

        c.set_interior_ring(i)
        self.assertTrue(c.has_interior_ring())
        self.assertIsInstance(c.get_interior_ring(), cfdm.InteriorRing)
        self.assertIsInstance(c.del_interior_ring(None), cfdm.InteriorRing)
        self.assertFalse(c.has_interior_ring())
        self.assertIsNone(c.del_interior_ring(None))

    def test_AuxiliaryCoordinate_hdf5_chunksizes(self):
        """Test the AuxiliaryCoordinate HDF5 chunksizes methods."""
        c = self.aux1.copy()
        i = cfdm.InteriorRing(data=cfdm.Data(np.arange(26).reshape(13, 2)))
        c.set_interior_ring(i)

        self.assertIsNone(c.nc_set_hdf5_chunksizes([4]))
        self.assertEqual(c.nc_hdf5_chunksizes(), (4,))
        self.assertEqual(c.bounds.nc_hdf5_chunksizes(), (4, 2))
        self.assertEqual(c.interior_ring.nc_hdf5_chunksizes(), (4, 2))

        c.nc_set_hdf5_chunksizes(1024, bounds=False, interior_ring=False)
        self.assertEqual(c.nc_hdf5_chunksizes(), 1024)
        self.assertEqual(c.bounds.nc_hdf5_chunksizes(), (4, 2))
        self.assertEqual(c.interior_ring.nc_hdf5_chunksizes(), (4, 2))

        c.nc_clear_hdf5_chunksizes(bounds=False, interior_ring=False)
        self.assertIsNone(c.nc_hdf5_chunksizes())
        self.assertEqual(c.bounds.nc_hdf5_chunksizes(), (4, 2))
        self.assertEqual(c.interior_ring.nc_hdf5_chunksizes(), (4, 2))

        c.nc_clear_hdf5_chunksizes()
        self.assertIsNone(c.nc_hdf5_chunksizes())
        self.assertIsNone(c.bounds.nc_hdf5_chunksizes())
        self.assertIsNone(c.interior_ring.nc_hdf5_chunksizes())


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



from .abstract import MeshArray
from .mixin import CompressedArrayMixin
from .subarray import (
    PointTopologyFromEdgesSubarray,
    PointTopologyFromFacesSubarray,
)


class PointTopologyArray(CompressedArrayMixin, MeshArray):
    """A point cell domain topology array derived from a UGRID variable.

    A point cell domain topology array derived from an underlying
    UGRID "edge_node_connectivity" or UGRID "face_node_connectivity"
    array.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def __new__(cls, *args, **kwargs):
        """Store subarray classes.

        If a child class requires different subarray classes than the
        ones defined here, then they must be defined in the __new__
        method of the child class.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        instance = super().__new__(cls)
        instance._Subarray = {
            "point topology from edges": PointTopologyFromEdgesSubarray,
            "point topology from faces": PointTopologyFromFacesSubarray,
        }
        return instance

    def __init__(
        self,
        edge_node_connectivity=None,
        face_node_connectivity=None,
        shape=None,
        start_index=None,
        cell_dimension=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            edge_node_connectivity: array_like, optional
                A 2-d integer array of indices that corresponds to a
                UGRID "edge_node_connectivity" variable.

            face_node_connectivity: array_like, optional
                A 2-d integer array of indices that corresponds to a
                UGRID "face_node_connectivity" variable.

            shape: `tuple`, optional
                The shape of the point cell domain topology array. If
                the shape is unknown (because the connectivity array
                has not been read yet) then set to `None`, which will
                result in a shape of ``(nan, nan)``.

            {{init start_index: `int`}}

            {{init cell_dimension: `int`}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        if edge_node_connectivity is not None:
            node_connectivity = edge_node_connectivity
            compression_type = "point topology from edges"
            if face_node_connectivity is not None:
                raise ValueError(
                    "Can't set both the 'edge_node_connectivity' and "
                    "'face_node_connectivity' parameters"
                )
        elif face_node_connectivity is not None:
            node_connectivity = face_node_connectivity
            compression_type = "point topology from faces"
        else:
            node_connectivity = None
            compression_type = None

        if shape is None:
            from math import nan

            shape = (nan, nan)

        # Note: Setting compressed_dimensions={0: (0,), 1: (1,)} means
        #       that only one subarray will ever span each of the two
        #       dimensions of the 'node_connectivity' array.
        super().__init__(
            connectivity=node_connectivity,
            shape=shape,
            compressed_dimensions={0: (0,), 1: (1,)},
            compression_type=compression_type,
            start_index=start_index,
            cell_dimension=cell_dimension,
            source=source,
            copy=copy,
        )



from . import CoordinateConversion, Datum, abstract


class CoordinateReference(abstract.Container):
    """A coordinate reference construct of the CF data model.

    A coordinate reference construct relates the coordinate values of
    the coordinate system to locations in a planetary reference frame.

    The domain of a field construct may contain various coordinate
    systems, each of which is constructed from a subset of the
    dimension and auxiliary coordinate constructs. For example, the
    domain of a four-dimensional field construct may contain
    horizontal (y-x), vertical (z), and temporal (t) coordinate
    systems. There may be more than one of each of these, if there is
    more than one coordinate construct applying to a particular
    spatiotemporal dimension (for example, there could be both
    latitude-longitude and y-x projection coordinate systems). In
    general, a coordinate system may be constructed implicitly from
    any subset of the coordinate constructs, yet a coordinate
    construct does not need to be explicitly or exclusively associated
    with any coordinate system.

    A coordinate system of the field construct can be explicitly
    defined by a coordinate reference construct which relates the
    coordinate values of the coordinate system to locations in a
    planetary reference frame and consists of the following:

    * References to the dimension coordinate and auxiliary coordinate
      constructs that define the coordinate system to which the
      coordinate reference construct applies. Note that the coordinate
      values are not relevant to the coordinate reference construct,
      only their properties.

    ..

    * A definition of a datum specifying the zeroes of the dimension
      and auxiliary coordinate constructs which define the coordinate
      system. The datum may be implied by the metadata of the
      referenced dimension and auxiliary coordinate constructs, or
      explicitly provided.

    ..

    * A coordinate conversion, which defines a formula for converting
      coordinate values taken from the dimension or auxiliary
      coordinate constructs to a different coordinate system. A
      coordinate reference construct relates the coordinate values of
      the field to locations in a planetary reference frame.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __new__(cls, *args, **kwargs):
        """Store component classes."""
        instance = super().__new__(cls)
        instance._CoordinateConversion = CoordinateConversion
        instance._Datum = Datum
        return instance

    def __init__(
        self,
        coordinates=None,
        datum=None,
        coordinate_conversion=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            coordinates: sequence of `str`, optional
                Identify the related dimension and auxiliary
                coordinate constructs by their construct
                identifiers.

                The coordinates may also be set after initialisation
                with the `set_coordinates` and `set_coordinate`
                methods.

                *Parameter example:*
                  ``coordinates=['dimensioncoordinate2']``

                *Parameter example:*
                  ``coordinates=('dimensioncoordinate0', 'dimensioncoordinate1')``

            datum: `Datum`, optional
                Set the datum component of the coordinate reference
                construct.

                The datum may also be set after initialisation with
                the `set_datum` method.

            coordinate_conversion: `CoordinateConversion`, optional
                Set the coordinate conversion component of the
                coordinate reference construct.

                The coordinate conversion may also be set after
                initialisation with the `set_coordinate_conversion`
                method.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

        self._set_component("coordinates", set(), copy=False)

        if source:
            try:
                coordinates = source.coordinates()
            except AttributeError:
                coordinates = None

            try:
                coordinate_conversion = source.get_coordinate_conversion()
            except AttributeError:
                coordinate_conversion = None

            try:
                datum = source.get_datum()
            except AttributeError:
                datum = None

        if coordinates is not None:
            self.set_coordinates(coordinates)

        if coordinate_conversion is not None:
            self.set_coordinate_conversion(coordinate_conversion, copy=copy)

        if datum is not None:
            self.set_datum(datum, copy=copy)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def construct_type(self):
        """Return a description of the construct type.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `str`
                The construct type.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.construct_type
        'coordinate_reference'

        """
        return "coordinate_reference"

    @property
    def coordinate_conversion(self):
        """Return the coordinate conversion component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `datum`, `get_coordinate_conversion`

        :Returns:

            `CoordinateConversion`
                The coordinate conversion.

        **Examples**

        >>> orog = {{package}}.DomainAncillary()
        >>> c = {{package}}.CoordinateConversion(
        ...     parameters={
        ...         'standard_name': 'atmosphere_hybrid_height_coordinate',
        ...     },
        ...     domain_ancillaries={'orog': orog}
        ... )
        >>> r = {{package}}.{{class}}(coordinate_conversion=c)
        >>> r.coordinate_conversion
        <{{repr}}CoordinateConversion: Parameters: standard_name; Ancillaries: orog>

        """
        return self.get_coordinate_conversion()

    @property
    def datum(self):
        """Return the datum component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `coordinate_conversion`, `get_datum`

        :Returns:

           `Datum`
                The datum.

        **Examples**

        >>> d = {{package}}.Datum(parameters={'earth_radius': 7000000})
        >>> r = {{package}}.{{class}}(datum=d)
        >>> r
        <{{repr}}CoordinateReference: >
        >>> r.datum
        <{{repr}}Datum: Parameters: earth_radius>

        """
        return self.get_datum()

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def clear_coordinates(self):
        """Remove all references to coordinate constructs.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_coordinate`, `coordinates`, `set_coordinates`

        :Returns:

            `set`
                The removed coordinate construct keys.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> r.set_coordinates(['dimensioncoordinate0', 'auxiliarycoordinate1'])
        >>> sorted(r.coordinates())
        ['auxiliarycoordinate1', 'dimensioncoordinate0']
        >>> sorted(r.clear_coordinates())
        ['auxiliarycoordinate1', 'dimensioncoordinate0']
        >>> r.coordinates()
        set()

        """
        out = self._get_component("coordinates")
        self._set_component("coordinates", set(), copy=False)
        return out.copy()

    def coordinates(self):
        """Return all references to coordinate constructs.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `clear_coordinates`, `set_coordinates`

        :Returns:

            `set`
                The coordinate construct keys.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> r.set_coordinates(['dimensioncoordinate0', 'auxiliarycoordinate1'])
        >>> sorted(r.coordinates())
        ['auxiliarycoordinate1', 'dimensioncoordinate0']
        >>> sorted(r.clear_coordinates())
        ['auxiliarycoordinate1', 'dimensioncoordinate0']
        >>> r.coordinates()
        set()

        """
        return self._get_component("coordinates").copy()

    def copy(self):
        """Return a deep copy.

        ``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `{{class}}`
                The deep copy.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> s = r.copy()

        """
        return type(self)(source=self, copy=True)

    def del_coordinate(self, key, default=ValueError()):
        """Remove a reference to a coordinate construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `coordinates`, `has_coordinate`, `set_coordinate`

        :Parameters:

            key: `str`
                The construct key of the coordinate construct.

                  *Parameter example:*
                     ``key='dimensioncoordinate1'``

                  *Parameter example:*
                     ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the
                coordinate construct has not been set. If set to an
                `Exception` instance then it will be raised instead.

        :Returns:

              The removed coordinate construct key.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> r.set_coordinates(['dimensioncoordinate0', 'auxiliarycoordinate1'])
        >>> r.coordinates()
        {'auxiliarycoordinate1', 'dimensioncoordinate0'}
        >>> r.del_coordinate('dimensioncoordinate0')
        'dimensioncoordinate0'
        >>> r.coordinates()
        {'auxiliarycoordinate1'}
        >>> r.del_coordinate('dimensioncoordinate0', 'not set')
        'not set'

        """
        coordinates = self._get_component("coordinates")
        if key in coordinates:
            coordinates.remove(key)
            return key

        if default is None:
            return

        return self._default(
            default, f"{self.__class__.__name__!r} has no {key!r} coordinate"
        )

    def del_coordinate_conversion(self):
        """Remove the coordinate conversion component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `coordinate_conversion`, `get_coordinate_conversion`,
                     `set_coordinate_conversion`

        :Returns:

            `CoordinateConversion`
                The removed coordinate conversion component.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> orog = {{package}}.DomainAncillary()
        >>> c = {{package}}.CoordinateConversion(
        ...     parameters={
        ...         'standard_name': 'atmosphere_hybrid_height_coordinate',
        ...     },
        ...     domain_ancillaries={'orog': orog}
        ... )
        >>> r.set_coordinate_conversion(c)
        >>> r.get_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: standard_name; Ancillaries: orog>
        >>> r.del_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: standard_name; Ancillaries: orog>
        >>> r.get_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: ; Ancillaries: >

        """
        new = self._CoordinateConversion()
        out = self._del_component("coordinate_conversion", new)
        self.set_coordinate_conversion(new, copy=False)
        return out

    def del_datum(self):
        """Remove the datum component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `datum`, `get_datum`, `set_datum`

        :Returns:

            `Datum`
                The removed datum component.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> d = {{package}}.Datum(parameters={'earth_radius': 7000000})
        >>> r.set_datum(d)
        >>> r.get_datum()
        <{{repr}}Datum: Parameters: earth_radius>
        >>> r.del_datum()
        <{{repr}}Datum: Parameters: earth_radius>
        >>> r.get_datum()
        <{{repr}}Datum: Parameters: >

        """
        new = self._Datum()
        out = self._del_component("datum", new)
        self.set_datum(new, copy=False)
        return out

    def get_coordinate_conversion(self):
        """Get the coordinate conversion component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `coordinate_conversion`, `del_coordinate_conversion`,
                     `set_coordinate_conversion`

        :Returns:

            `CoordinateConversion`
                The coordinate conversion component.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> orog = {{package}}.DomainAncillary()
        >>> c = {{package}}.CoordinateConversion(
        ...     parameters={
        ...         'standard_name': 'atmosphere_hybrid_height_coordinate',
        ...     },
        ...     domain_ancillaries={'orog': orog}
        ... )
        >>> r.set_coordinate_conversion(c)
        >>> r.get_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: standard_name; Ancillaries: orog>
        >>> r.del_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: standard_name; Ancillaries: orog>
        >>> r.get_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: ; Ancillaries: >

        """
        out = self._get_component("coordinate_conversion", None)
        if out is None:
            out = self._CoordinateConversion()
            self.set_coordinate_conversion(out, copy=False)

        return out

    def get_datum(self):
        """Return the datum component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `datum`, `get_coordinate_conversion`

        :Returns:

           `Datum`
                The datum component.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> d = {{package}}.Datum(parameters={'earth_radius': 7000000})
        >>> r.set_datum(d)
        >>> r.get_datum()
        <{{repr}}Datum: Parameters: earth_radius>
        >>> r.del_datum()
        <{{repr}}Datum: Parameters: earth_radius>
        >>> r.get_datum()
        <{{repr}}Datum: Parameters: >

        """
        out = self._get_component("datum", None)
        if out is None:
            out = self._Datum()
            self.set_datum(out, copy=False)

        return out

    def has_coordinate(self, key):
        """Whether a reference to a coordinate construct has been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `coordinates`, `del_coordinate`, `set_coordinate`

        :Parameters:

            key: `str`
                The construct key of the coordinate construct.

                  *Parameter example:*
                     ``key='dimensioncoordinate1'``

                  *Parameter example:*
                     ``key='auxiliarycoordinate0'``

        :Returns:

              `bool`
                  `True` if the coordinate construct key has been set,
                  otherwise `False`.

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> r.set_coordinates(['dimensioncoordinate0', 'auxiliarycoordinate1'])
        >>> r.coordinates()
        {'auxiliarycoordinate1', 'dimensioncoordinate0'}
        >>> r.has_coordinate('dimensioncoordinate0')
        True
        >>> r.has_coordinate('dimensioncoordinate1')
        False

        """
        return key in self._get_component("coordinates")

    def set_coordinate(self, key):
        """Set a reference to a coordinate construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `coordinates`, `del_coordinate`, `has_coordinate`

        :Parameters:

            key: `str`
                The construct key of the coordinate construct.

                  *Parameter example:*
                     ``key='dimensioncoordinate1'``

                  *Parameter example:*
                     ``key='auxiliarycoordinate0'``

        :Returns:

            `None`

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> r.set_coordinate('dimensioncoordinate0')
        >>> r.coordinates()
        {'dimensioncoordinate0'}
        >>> r.has_coordinate('dimensioncoordinate0')
        True
        >>> r.has_coordinate('auxiliarycoordinate0')
        False

        """
        c = self._get_component("coordinates")
        c.add(key)

    def set_coordinates(self, coordinates):  # SB NOTE: flaky doctest set order
        """Set references to coordinate constructs.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `coordinates`, `clear_coordinates`, `set_coordinate`

        :Parameters:

            coordinates: (sequence of) `str`
                The coordinate construct keys to be set.

                *Parameter example:*
                  ``coordinates=['dimensioncoordinate0']``

                *Parameter example:*
                  ``coordinates='dimensioncoordinate0'``

                *Parameter example:*
                  ``coordinates=set(['dimensioncoordinate0', 'auxiliarycoordinate1'])``

                *Parameter example:*
                  ``coordinates=[]``

        :Returns:

            `None`

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> r.set_coordinates(['dimensioncoordinate0', 'auxiliarycoordinate1'])
        >>> r.coordinates()
        {'auxiliarycoordinate1', 'dimensioncoordinate0'}
        >>> r.clear_coordinates()
        {'auxiliarycoordinate1', 'dimensioncoordinate0'}
        >>> r.coordinates()
        set()

        """
        if isinstance(coordinates, str):
            coordinates = (coordinates,)

        self._get_component("coordinates").update(coordinates)

    def set_coordinate_conversion(self, coordinate_conversion, copy=True):
        """Set the coordinate conversion component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `coordinate_conversion`, `del_coordinate_conversion`,
                     `get_coordinate_conversion`

        :Parameters:

            coordinate_conversion: `CoordinateConversion`
                The coordinate conversion component to be inserted.

            {{copy: `bool`, optional}}

        :Returns:

            `None`

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> orog = {{package}}.DomainAncillary()
        >>> c = {{package}}.CoordinateConversion(
        ...     parameters={
        ...         'standard_name': 'atmosphere_hybrid_height_coordinate',
        ...     },
        ...     domain_ancillaries={'orog': orog}
        ... )
        >>> r.set_coordinate_conversion(c)
        >>> r.get_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: standard_name; Ancillaries: orog>
        >>> r.del_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: standard_name; Ancillaries: orog>
        >>> r.get_coordinate_conversion()
        <{{repr}}CoordinateConversion: Parameters: ; Ancillaries: >

        """
        if copy:
            coordinate_conversion = coordinate_conversion.copy()

        self._set_component(
            "coordinate_conversion", coordinate_conversion, copy=False
        )

    def set_datum(self, datum, copy=True):
        """Set the datum component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `datum`, `del_datum`, `get_datum`

        :Parameters:

            datum: `Datum`
                The datum component to be inserted.

            {{copy: `bool`, optional}}

        :Returns:

            `None`

        **Examples**

        >>> r = {{package}}.{{class}}()
        >>> d = {{package}}.Datum(parameters={'earth_radius': 7000000})
        >>> r.set_datum(d)
        >>> r.get_datum()
        <{{repr}}Datum: Parameters: earth_radius>
        >>> r.del_datum()
        <{{repr}}Datum: Parameters: earth_radius>
        >>> r.get_datum()
        <{{repr}}Datum: Parameters: >

        """
        if copy:
            datum = datum.copy()

        self._set_component("datum", datum, copy=False)
